/* holoCryptoFourier.c
 *
 * Copyright (C) 2016, 2017 Nicolas Pavillon.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

// flomar, 2017/03/13 (be): This source file has been modified slightly
// to be compilable under Linux, MacOS, and Windows.
// Tested under Linux (gcc), MacOS (gcc), and Windows 7 (VS2015). 
//
// Compilation under Linux: gcc holoCryptoFourier.c -o holoCryptoFourier -std=gnu99 -lm
// Compilation under MacOS: gcc holoCryptoFourier.c -o holoCryptoFourier
// Compilation under Windows: cl holoCryptoFourier.c /Fe:holoCryptoFourier.exe
//
// Please uncomment the lines below (simply remove the leading "//") to use 
// the corresponding libraries for improved performance. Consult the README 
// file for build instructions when these macros are active.

//#define USEFFTW 1
//#define USETIFF 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef _WIN32
#define _USE_MATH_DEFINES
#include <math.h>
#include <time.h>
#elif __linux__
#include <math.h>
#include <time.h>
#else
#include <math.h>
#include <sys/time.h>
#endif

#ifdef USEFFTW
#include <fftw3.h>
#else
//Internal definition for FFT, heavily inspired by the FFT code in GSL
typedef double fftw_complex[2];
int checkPow2(int dataSize);
int intern2DFFT(fftw_complex* data, int dataSize, int sign);
int bitreverse_order2D(fftw_complex* data, int dataSize);
#endif
#ifdef USETIFF
#include <tiffio.h>
//Functions using tiff library for writing images
int tiffGetImageSize(char* filename, int* imageHeight, int* imageWidth);
int tiffReadImage(char* filename, int imageHeight, int imageWidth, unsigned short* imData);
int tiffWriteImage(char* filename, int imageHeight, int imageWidth, unsigned char bitsPerPixel, unsigned short* imData);
#else
//Functions handling I/O of csv files
int csvGetImageSize(char* filename, int* imageHeight, int* imageWidth);
int csvReadImage(char* filename, int imageHeight, int imageWidth, unsigned short* imData);
int csvWriteImage(char* filename, int imageHeight, int imageWidth, unsigned short* imData);
#endif

const char version[]="holoCryptoFourier v1.0.1_2017-07-08";

//Macro to access real part of a complex number stored as defined in the FFTW
#define REAL(a,i) ((a)[(i)][0])
//Macro to access imaginary part of a complex number stored as defined in the FFTW
#define IMAG(a,i) ((a)[(i)][1])
//Fixed length of file names
#define FILELENGTH 200
//Fixed length of line length in files
#define LINELENGTH 5000

enum { ENCRYPT, DECRYPT, GENERATE };

//Functions for input/output
int getImageSize(char* filename, int* imageHeight, int* imageWidth);
int readImage(char* filename, int imageHeight, int imageWidth, unsigned short* imData);
int writeImage(char* filename, int imageHeight, int imageWidth, unsigned char bitsPerPixel, unsigned short* imData);

//Generic functions for manipulation of complex numbers
int complexMult(fftw_complex* data, fftw_complex* data2, int imageHeight, int imageWidth, fftw_complex* dataOut);
int complexMultConj(fftw_complex* data, fftw_complex* data2, int imageHeight, int imageWidth, fftw_complex* dataOut);
int polar2Cart(fftw_complex* data, int imageHeight, int imageWidth);
int cart2Polar(fftw_complex* data, int imageHeight, int imageWidth);

//Main function for holographic encryption
int readMasks(char* mask1File, char* mask2File,
              fftw_complex *mask1, fftw_complex *mask2, int imageHeight, int imageWidth);
int encryptHolo(char* inputFile, char* outputFile, char* mask1File, char* mask2File, int imageHeight, int imageWidth);
int decryptHolo(char* inputFile, char* outputFile, char* mask1File, char* mask2File, int imageHeight, int imageWidth);
int generateMasks(char *mask1File, char *mask2File, int imageHeight, int imageWidth);

int readArguments(int argc, const char* argv[], int* mode, char* inputFile, char* outputFile,
                  char* mask1File, char* mask2File, int* imageHeight, int* imageWidth);

//Amount of levels in phase masks
const int maskLevel = 255;

int main(int argc, const char* argv[]) {

    char inputFile[FILELENGTH];
    char outputFile[FILELENGTH];
    char mask1File[FILELENGTH];
    char mask2File[FILELENGTH];
    
    int res, imageHeight, imageWidth, mode;
    
    //Extract arguments from command line
    res = readArguments(argc, argv, &mode, inputFile, outputFile, mask1File, mask2File, &imageHeight, &imageWidth);
    if (res)
        return res;
    
    switch (mode) {
        case ENCRYPT:
            printf("Encrypting image %s.\n", inputFile);
            encryptHolo(inputFile, outputFile, mask1File, mask2File, imageHeight, imageWidth);
            break;
        case DECRYPT:
            printf("Decrypting image.\n");
            decryptHolo(inputFile, outputFile, mask1File, mask2File, imageHeight, imageWidth);
            break;
        case GENERATE:
            printf("Generating phase masks of size %d x %d for encryption of file %s.\n",
                   imageHeight, imageWidth, inputFile);
            generateMasks(mask1File, mask2File, imageHeight, imageWidth);
            break;
        default:
            printf("Unknown mode, quitting");
            return 0;
            break;
    }
    
    return 0;
    
}

int readArguments(int argc, const char *argv[], int *mode, char *inputFile, char *outputFile,
                  char *mask1File, char *mask2File, int *imageHeight, int *imageWidth) {
    if (argv == NULL || mode == NULL || inputFile == NULL || outputFile == NULL ||
        mask1File == NULL || mask2File == NULL || imageHeight == NULL || imageWidth == NULL) {
        printf("readArguments: null pointer.\n");
        return 1;
    }
    int res, width, height, k;
    char filename[FILELENGTH];
#ifdef USETIFF
    char extension[] = ".tif";
#else
    char extension[] = ".csv";
#endif
    
    if (argc < 2)
        goto usage;
    
    //Read first argument for mode
    switch (argv[1][0]) {
        case 'e':
            *mode = ENCRYPT;
            if (argc != 6) {
                printf("Error: incorrect amount of arguments.\n");
                goto usage;
            }
            break;
        case 'd':
            *mode = DECRYPT;
            if (argc != 6) {
                printf("Error: incorrect amount of arguments.\n");
                goto usage;
            }
            break;
        case 'g':
            *mode = GENERATE;
            if (argc != 5) {
                printf("Error: incorrect amount of arguments.\n");
                goto usage;
            }
            break;
        default:
            printf("Unknown mode.\n");
            goto usage;
            break;
    }
    
    //Get mask file names
    sprintf(mask1File, "%s%s", argv[2], extension);
    sprintf(mask2File, "%s%s", argv[3], extension);
    
    //Continue checks by open input image first
    if (*mode == ENCRYPT || *mode == GENERATE) {
        strcpy(inputFile, argv[4]);
        sprintf(filename, "%s0%s", inputFile, extension);
        res = getImageSize(filename, imageHeight, imageWidth);
        if (res)
            return res;
        if (*imageHeight < 1 || *imageWidth < 1) {
            printf("Error: width and height of images must be > 1.\n");
            return 2;
        }
#ifndef USEFFTW
        if (checkPow2(*imageHeight)) {
            printf("Error: image size must be a power of 2.\n");
            return 2;
        }
        if (checkPow2(*imageWidth)) {
            printf("Error: image size must be a power of 2.\n");
            return 2;
        }
#endif
        if (*mode == ENCRYPT) {
            sprintf(filename, "%s1%s", inputFile, extension);
            res = getImageSize(filename, &height, &width);
            if (res)
                return res;
            if (height != *imageHeight || width != *imageWidth)
                printf("Images have different sizes.\n");
            strcpy(outputFile, argv[5]);
        }
    }
    else {
        sprintf(filename, "%s0%s", argv[4], extension);
        res = getImageSize(filename, imageHeight, imageWidth);
        if (res)
            return res;
#ifndef USEFFTW
        if (checkPow2(*imageHeight)) {
            printf("Error: image size must be a power of 2.\n");
            return 2;
        }
        if (checkPow2(*imageWidth)) {
            printf("Error: image size must be a power of 2.\n");
            return 2;
        }
#endif
        for (k = 1; k < 3; k++) {
            sprintf(filename, "%s%d%s", argv[4], k, extension);
            res = getImageSize(filename, &height, &width);
            if (res)
                return res;
            if (height != *imageHeight || width != *imageWidth)
                printf("Size of images are different.\n");
        }
        strcpy(inputFile, argv[4]);
        strcpy(outputFile, argv[5]);
    }
    if (*mode == ENCRYPT || *mode == DECRYPT) {
        res = getImageSize(mask1File, &height, &width);
        if (res)
            return res;
        if (height != *imageHeight || width != *imageWidth)
            printf("Size of mask1 is different from image.\n");
        res = getImageSize(mask2File, &height, &width);
        if (res)
            return res;
        if (height != *imageHeight || width != *imageWidth)
            printf("Size of mask2 is different from image.\n");
    }

    return 0;

usage:
    printf("Usage of program (%s):\n", version);

    printf("Encryption:\n    holoCryptoFourier e mask1 mask2 input output\n");
    printf("Decryption:\n    holoCryptoFourier d mask1 mask2 input output\n");
    printf("Generation of masks:\n    holoCryptoFourier g mask1 mask2 input\n");
    printf("where mask1, mask2, input, output are names without extensions.\n");
    printf("For encryption: reads input0.tif and input1.tif, and saves output0.tif, ..., output2.tif.\n");
    printf("For decryption: reads input0.tif, ..., input2.tif, and saves output0.tif and output1.tif\n");
                              
    return 1;
}

int encryptHolo(char *inputFile, char *outputFile, char *mask1File, char *mask2File, int imageHeight, int imageWidth) {
    if (inputFile == NULL || outputFile == NULL || mask1File == NULL || mask2File == NULL) {
        printf("encryptHolo: null pointer.\n");
        return 1;
    }
    
#ifdef USETIFF
    char extension[] = ".tif";
#else
    char extension[] = ".csv";
#endif
    unsigned short *holo = (unsigned short*)malloc(4*imageHeight*imageWidth*sizeof(unsigned short));
    double *intensity = (double*)malloc(4*imageHeight*imageWidth*sizeof(double));
    
#ifdef USEFFTW
    fftw_complex *mask1 = (fftw_complex*)fftw_malloc(imageHeight*imageWidth*sizeof(fftw_complex));
    fftw_complex *mask2 = (fftw_complex*)fftw_malloc(imageHeight*imageWidth*sizeof(fftw_complex));
    fftw_complex *data = (fftw_complex*)fftw_malloc(imageHeight*imageWidth*sizeof(fftw_complex));
#else
    fftw_complex *mask1 = (fftw_complex*)malloc(imageHeight*imageWidth*sizeof(fftw_complex));
    fftw_complex *mask2 = (fftw_complex*)malloc(imageHeight*imageWidth*sizeof(fftw_complex));
    fftw_complex *data = (fftw_complex*)malloc(imageHeight*imageWidth*sizeof(fftw_complex));
#endif
    fftw_complex value, ref;
    
    if (holo == NULL || intensity == NULL || mask1 == NULL || mask2 == NULL || data == NULL) {
        if (holo != NULL) free(holo);
        if (intensity != NULL) free(intensity);
        if (mask1 != NULL) free(mask1);
        if (mask2 != NULL) free(mask2);
        if (data != NULL) free(data);
        printf("encryptHolo: error in memory allocation.\n");
        return 1;
    }
    
    int res, shift, k;
    double meanVal, maxVal = 0;
    char holoName[FILELENGTH];
    
#ifdef USEFFTW
    //Preparing basic FFTW plan
    fftw_plan planComplexForward = fftw_plan_dft_2d(imageHeight, imageWidth, data, data, FFTW_FORWARD, FFTW_ESTIMATE);
    fftw_plan planComplexBackward = fftw_plan_dft_2d(imageHeight, imageWidth, data, data, FFTW_BACKWARD, FFTW_ESTIMATE);
#endif
    
    //Prepare masks
    res = readMasks(mask1File, mask2File, mask1, mask2, imageHeight, imageWidth);
    if (res)
        goto encryptCleanMem;
    
    //Reading image
    sprintf(holoName, "%s0%s", inputFile, extension);
    res = readImage(holoName, imageHeight, imageWidth, holo);
    if (res)
        goto encryptCleanMem;
    //Copy to amplitude with normalization
    for (k = 0; k < imageHeight*imageWidth; k++)
        REAL(data, k) = holo[k]/255.;
    
    //Reading second image
    sprintf(holoName, "%s1%s", inputFile, extension);
    res = readImage(holoName, imageHeight, imageWidth, holo);
    if (res)
        goto encryptCleanMem;
    //Copy to phase with scaling
    for (k = 0; k < imageHeight*imageWidth; k++)
        IMAG(data, k) = holo[k]/256.*2*M_PI;
    
    //Convert to Cartesian
    polar2Cart(data, imageHeight, imageWidth);
    
    //Multiplying by mask1
    complexMult(data, mask1, imageHeight, imageWidth, data);
    
    //In-place FFT
#ifdef USEFFTW
    fftw_execute(planComplexForward);
#else
    //-1 means exp(-i omega x), i.e. forward FFT
    intern2DFFT(data, imageHeight, -1);
#endif
    //Multiplying by mask2
    complexMult(data, mask2, imageHeight, imageWidth, data);
    
    //In-place iFFT
#ifdef USEFFTW
    fftw_execute(planComplexBackward);
#else
    intern2DFFT(data, imageHeight, 1);
#endif
    //Normalization
    for (k = 0; k < imageHeight*imageWidth; k++) {
        REAL(data, k) /= imageHeight*imageWidth;
        IMAG(data, k) /= imageHeight*imageWidth;
    }
    
    //Compute mean value of object
    meanVal = 0;
    for (k = 0; k < imageHeight*imageWidth; k++)
        meanVal += sqrt( REAL(data, k)*REAL(data, k) + IMAG(data, k)*IMAG(data, k) );
    meanVal /= imageHeight*imageWidth;
    
    maxVal = 0;
    //Perform interference with flat reference
    for (shift = 0; shift < 3; shift++) {
        //Prepare reference value with chosen phase shift
        ref[0] = meanVal*cos(shift*M_PI/3.);
        ref[1] = meanVal*sin(shift*M_PI/3.);
        
        //Add reference
        for (k = 0; k < imageHeight*imageWidth; k++) {
            value[0] = REAL(data, k) + ref[0];
            value[1] = IMAG(data, k) + ref[1];
            
            intensity[shift*imageHeight*imageWidth + k] = value[0]*value[0] + value[1]*value[1];
            
            //Check max value in images for normalization
            if (intensity[shift*imageHeight*imageWidth + k] > maxVal)
                maxVal = intensity[shift*imageHeight*imageWidth + k];
        }
    }
    
    for (shift = 0; shift < 3; shift++) {
        //Scaling for 16-bit images
        for (k = 0; k < imageHeight*imageWidth; k++)
            holo[k] = (unsigned short)(intensity[shift*imageHeight*imageWidth + k]/maxVal*65535);
        
        sprintf(holoName, "%s%d%s", outputFile, shift, extension);
        
        res = writeImage(holoName, imageHeight, imageWidth, 16, holo);
        if (res)
            goto encryptCleanMem;
    }
    
encryptCleanMem:
    free(holo);
    free(intensity);
#ifdef USEFFTW
    fftw_free(mask1);
    fftw_free(mask2);
    fftw_free(data);
    fftw_destroy_plan(planComplexForward);
    fftw_destroy_plan(planComplexBackward);
#else
    free(mask1);
    free(mask2);
    free(data);
#endif
    return res;
}

int decryptHolo(char *inputFile, char *outputFile, char *mask1File, char *mask2File, int imageHeight, int imageWidth) {
    if (inputFile == NULL || outputFile == NULL || mask1File == NULL || mask2File == NULL) {
        printf("encryptHolo: null pointer.\n");
        return 1;
    }
    
    int shift, res, k;
    double maxVal, num, denom;
    char holoName[FILELENGTH];
#ifdef USETIFF
    char extension[] = ".tif";
#else
    char extension[] = ".csv";
#endif
    unsigned short *holo = (unsigned short*)malloc(4*imageHeight*imageWidth*sizeof(unsigned short));
#ifdef USEFFTW
    fftw_complex *mask1 = (fftw_complex*)fftw_malloc(imageHeight*imageWidth*sizeof(fftw_complex));
    fftw_complex *mask2 = (fftw_complex*)fftw_malloc(imageHeight*imageWidth*sizeof(fftw_complex));
    fftw_complex *data = (fftw_complex*)fftw_malloc(imageHeight*imageWidth*sizeof(fftw_complex));
#else
    fftw_complex *mask1 = (fftw_complex*)malloc(imageHeight*imageWidth*sizeof(fftw_complex));
    fftw_complex *mask2 = (fftw_complex*)malloc(imageHeight*imageWidth*sizeof(fftw_complex));
    fftw_complex *data = (fftw_complex*)malloc(imageHeight*imageWidth*sizeof(fftw_complex));
#endif
    if (holo == NULL || mask1 == NULL || mask2 == NULL || data == NULL) {
        if (holo != NULL) free(holo);
        if (mask1 != NULL) free(mask1);
        if (mask2 != NULL) free(mask2);
        if (data != NULL) free(data);
        printf("encryptHolo: error in memory allocation.\n");
        return 1;
    }
    
#ifdef USEFFTW
    //Preparing basic FFTW plan
    fftw_plan planComplexForward = fftw_plan_dft_2d(imageHeight, imageWidth, data, data, FFTW_FORWARD, FFTW_ESTIMATE);
    fftw_plan planComplexBackward = fftw_plan_dft_2d(imageHeight, imageWidth, data, data, FFTW_BACKWARD, FFTW_ESTIMATE);
#endif
    
    //Prepare masks
    res = readMasks(mask1File, mask2File, mask1, mask2, imageHeight, imageWidth);
    if (res)
        goto decryptCleanMem;
    
    //Reading the holograms
    for (shift = 0; shift < 3; shift++) {
        sprintf(holoName, "%s%d%s", inputFile, shift, extension);
        
        readImage(holoName, imageHeight, imageWidth, holo + shift*imageHeight*imageWidth);
    }
    //Retrieve field through phase-shifting
    for (k = 0; k < imageHeight*imageWidth; k++) {
        num = 2.*holo[k] - 3.*holo[imageHeight*imageWidth + k] + holo[2*imageHeight*imageWidth + k];
        denom = sqrt(3)* ((double)holo[imageHeight*imageWidth + k] - holo[2*imageHeight*imageWidth + k]);
        REAL(data, k) = sqrt(num*num + denom*denom);
        
        IMAG(data, k) = -atan2(num, denom);
    }
    
    //Convert to Cartesian
    polar2Cart(data, imageHeight, imageWidth);
    
    //In-place FFT
#ifdef USEFFTW
    fftw_execute(planComplexForward);
#else
    //-1 means exp(-i omega x), i.e. forward FFT
    intern2DFFT(data, imageHeight, -1);
#endif
    //Multiplying by mask2
    complexMultConj(data, mask2, imageHeight, imageWidth, data);
    
    //In-place iFFT
#ifdef USEFFTW
    fftw_execute(planComplexBackward);
#else
    intern2DFFT(data, imageHeight, 1);
#endif
    //Normalization
    for (k = 0; k < imageHeight*imageWidth; k++) {
        REAL(data, k) /= imageHeight*imageWidth;
        IMAG(data, k) /= imageHeight*imageWidth;
    }
    
    //Multiply by mask1
    complexMultConj(data, mask1, imageHeight, imageWidth, data);
    
    //Go back to polar for saving output
    cart2Polar(data, imageHeight, imageWidth);
    
    //One pass to find max value
    maxVal = 0;
    for (k = 0; k < imageHeight*imageWidth; k++) {
        if (REAL(data, k) > maxVal)
            maxVal = REAL(data, k);
    }
    //Copying data with scaling
    for (k = 0; k < imageHeight*imageWidth; k++)
        holo[k] = (unsigned char) round(REAL(data, k)/maxVal*255);
    //Saving amplitude image
    sprintf(holoName, "%s0%s", outputFile, extension);
    res = writeImage(holoName, imageHeight, imageWidth, 8, holo);
    if (res)
        goto decryptCleanMem;

    //Copy data of phase within standard bounds
    for (k = 0; k < imageHeight*imageWidth; k++)
        holo[k] = (unsigned char) ( (IMAG(data, k)/(2*M_PI) + 0.5)*255);
    //Saving phase image
    sprintf(holoName, "%s1%s", outputFile, extension);
    res = writeImage(holoName, imageHeight, imageWidth, 8, holo);
    
decryptCleanMem:
    free(holo);
#ifdef USEFFTW
    fftw_free(mask1);
    fftw_free(mask2);
    fftw_free(data);
    fftw_destroy_plan(planComplexForward);
    fftw_destroy_plan(planComplexBackward);
#else
    free(mask1);
    free(mask2);
    free(data);
#endif
    return res;
}

int generateMasks(char *mask1File, char *mask2File, int imageHeight, int imageWidth) {
    if (mask1File == NULL || mask2File == NULL) {
        printf("generateMasks: null pointer.\n");
        return 1;
    }
    
    unsigned short *mask = (unsigned short*)malloc(imageHeight*imageWidth*sizeof(unsigned short));
    if (mask == NULL) {
        printf("generateMasks: error in memory allocation.\n");
        return 1;
    }
    int res;
    srand((unsigned int)time(NULL));
    
    //Scanning on mask and generating values
    for (int k = 0; k < imageHeight*imageWidth; k++) {
        mask[k] = rand() % maskLevel;
    }
    res = writeImage(mask1File, imageHeight, imageWidth, 8, mask);
    if (res) {
        free(mask);
        return res;
    }
    
    //Generating second mask
    for (int k = 0; k < imageHeight*imageWidth; k++) {
        mask[k] = rand() % maskLevel;
    }
    res = writeImage(mask2File, imageHeight, imageWidth, 8, mask);
    
    free(mask);
    
    return res;
}

int readMasks(char *mask1File, char* mask2File, 
              fftw_complex *mask1, fftw_complex *mask2, int imageHeight, int imageWidth) {
    if (mask1File == NULL || mask2File == NULL || mask1 == NULL || mask2 == NULL) {
        printf("readMasks: null pointer.\n");
        return 1;
    }
    unsigned short *image = (unsigned short*)malloc(imageHeight*imageWidth*sizeof(unsigned short));
    if (image == NULL) {
        printf("readMasks: error in memory allocation.\n");
        return 1;
    }
    int res;
    
    res = readImage(mask1File, imageHeight, imageWidth, image);
    if (res)
        return res;
    //Transferring mask values to phase
    for (int k = 0; k < imageHeight*imageWidth; k++) {
        REAL(mask1, k) = 1;
        IMAG(mask1, k) = (double)image[k]/maskLevel * 2 * M_PI;
    }
    //Converting to Cartesian
    polar2Cart(mask1, imageHeight, imageWidth);
    
    res = readImage(mask2File, imageHeight, imageWidth, image);
    if (res)
        return res;
    //Transferring mask values to phase
    for (int k = 0; k < imageHeight*imageWidth; k++) {
        REAL(mask2, k) = 1;
        IMAG(mask2, k) = (double)image[k]/maskLevel * 2 * M_PI;
    }
    //Converting to Cartesian
    polar2Cart(mask2, imageHeight, imageWidth);
    
    free(image);
    
    return 0;
}

int complexMult(fftw_complex* data, fftw_complex* data2, int imageHeight, int imageWidth, fftw_complex* dataOut) {
    
    fftw_complex value;
    for (int k = 0; k < imageHeight*imageWidth; k++) {
        value[0] = REAL(data, k)*REAL(data2, k) - IMAG(data, k)*IMAG(data2, k);
        value[1] = IMAG(data, k)*REAL(data2, k) + REAL(data, k)*IMAG(data2, k);
        REAL(dataOut, k) = value[0];
        IMAG(dataOut, k) = value[1];
    }
    
    return 0;
}

int complexMultConj(fftw_complex* data, fftw_complex* data2, int imageHeight, int imageWidth, fftw_complex* dataOut) {
    
    fftw_complex value;
    for (int k = 0; k < imageHeight*imageWidth; k++) {
        value[0] = REAL(data, k)*REAL(data2, k) + IMAG(data, k)*IMAG(data2, k);
        value[1] = IMAG(data, k)*REAL(data2, k) - REAL(data, k)*IMAG(data2, k);
        REAL(dataOut, k) = value[0];
        IMAG(dataOut, k) = value[1];
    }
    
    return 0;
}


int cart2Polar(fftw_complex* data, int imageHeight, int imageWidth) {
    if (data == NULL) {
        printf("cart2Polar: null pointer.\n");
        return 1;
    }
    if (imageHeight < 1 || imageWidth < 1) {
        printf("cart2Polar: width and height must be > 1.\n");
        return 2;
    }
    
    int k;
    //Converting to amplitude and phase
    for (k = 0; k < imageHeight*imageWidth; k++) {
        double real = REAL(data, k);
        double imag = IMAG(data, k);
        REAL(data, k) = sqrt(real*real + imag*imag );
        IMAG(data, k) = atan2(imag, real);
    }
    
    return 0;
}

int polar2Cart(fftw_complex* data, int imageHeight, int imageWidth) {
    if (data == NULL) {
        printf("polar2Cart: null pointer.\n");
        return 1;
    }
    if (imageHeight < 1 || imageWidth < 1) {
        printf("polar2Cart: width and height must be > 1.\n");
        return 2;
    }
    
    int k;
    //Converting to real and imaginary
    for (k = 0; k < imageHeight*imageWidth; k++) {
        double abs = REAL(data, k);
        double phase = IMAG(data, k);
        REAL(data, k) = abs*cos(phase);
        IMAG(data, k) = abs*sin(phase);
    }
    
    return 0;
}

int getImageSize(char* filename, int* imageHeight, int* imageWidth) {
#ifdef USETIFF
    return tiffGetImageSize(filename, imageHeight, imageWidth);
#else
    return csvGetImageSize(filename, imageHeight, imageWidth);
#endif
}

int readImage(char* filename, int imageHeight, int imageWidth, unsigned short* imData) {
#ifdef USETIFF
    return tiffReadImage(filename, imageHeight, imageWidth, imData);
#else
    return csvReadImage(filename, imageHeight, imageWidth, imData);
#endif
}

int writeImage(char* filename, int imageHeight, int imageWidth,
               unsigned char bitsPerPixel, unsigned short* imData) {
#ifdef USETIFF
    return tiffWriteImage(filename, imageHeight, imageWidth, bitsPerPixel, imData);
#else
    return csvWriteImage(filename, imageHeight, imageWidth, imData);
#endif
}

#ifdef USETIFF
int tiffGetImageSize(char *filename, int* imageHeight, int* imageWidth) {
    if (filename == NULL || imageHeight == NULL || imageWidth == NULL) {
        printf("tiffGetImageSize: null input pointer.\n");
        return 1;
    }
    TIFFSetWarningHandler(NULL);
    TIFFSetErrorHandler(NULL);
    
    //Starting with null sizes
    *imageHeight = 0;
    *imageWidth = 0;
    
    TIFF* holoImage = TIFFOpen(filename, "r");
    if (holoImage == NULL) {
        printf("wrapperTiffGetImageSize: impossible to open file %s.\n", filename);
        return 2;
    }
    
    //Checking samples per pixel
    uint16 bitPerSample, samplePerPixel;
    if (!TIFFGetField(holoImage, TIFFTAG_BITSPERSAMPLE, &bitPerSample)) {
        TIFFClose(holoImage);
        printf("wrapperTiffGetImageSize: error while reading TIFF tag.\n");
        return 2;
    }
    if (!TIFFGetField(holoImage, TIFFTAG_SAMPLESPERPIXEL, &samplePerPixel)) {
        TIFFClose(holoImage);
        printf("wrapperTiffGetImageSize: error while reading TIFF tag.\n");
        return 2;
    }
    if (samplePerPixel != 1 || (bitPerSample != 8 && bitPerSample != 16)) {
        TIFFClose(holoImage);
        printf("wrapperTiffGetImageSize: unsupported bit depth.\n");
        return 2;
    }
    
    //Reading width and length
    uint32 width, length;
    if (!TIFFGetField(holoImage, TIFFTAG_IMAGEWIDTH, &width)) {
        TIFFClose(holoImage);
        printf("wrapperTiffGetImageSize: error while reading TIFF tag.\n");
        return 2;
    }
    if (!TIFFGetField(holoImage, TIFFTAG_IMAGELENGTH, &length)) {
        TIFFClose(holoImage);
        printf("wrapperTiffGetImageSize: error while reading TIFF tag.\n");
        return 2;
    }
    TIFFClose(holoImage);
    
    //Transferring to user variables
    *imageWidth = (int)width;
    *imageHeight = (int)length;
    
    return 0;
}

int tiffReadImage(char* filename, int imageHeight, int imageWidth, unsigned short* imData) {
    if (filename == NULL || imData == NULL) {
        printf("wrapperTiffReadImage: null input pointer.\n");
        return 1;
    }
    if (imageHeight < 1 || imageWidth < 1) {
        printf("wrapperTiffReadImage: width and height must be > 1.\n");
        return 2;
    }
    TIFFSetWarningHandler(NULL);
    TIFFSetErrorHandler(NULL);
    
    TIFF* holoImage = TIFFOpen(filename, "r");
    if (holoImage == NULL) {
        printf("wrapperTiffReadImage: impossible to open file.\n");
        return 2;
    }
    
    unsigned short samplePerPixel = 0, bitsPerPixel = 0;
    unsigned int width = 0, length = 0;
    
    int status = TIFFGetField(holoImage, TIFFTAG_BITSPERSAMPLE, &bitsPerPixel);
    if (status)
        status = TIFFGetField(holoImage, TIFFTAG_SAMPLESPERPIXEL, &samplePerPixel);
    if (status)
        status = TIFFGetField(holoImage, TIFFTAG_IMAGEWIDTH, &width);
    if (status)
        status = TIFFGetField(holoImage, TIFFTAG_IMAGELENGTH, &length);
    if (!status) {
        TIFFClose(holoImage);
        printf("wrapperTiffReadImage: error while reading TIFF tag.\n");
        return 2;
    }
    if (samplePerPixel != 1 || (bitsPerPixel != 8 && bitsPerPixel != 16)) {
        TIFFClose(holoImage);
        printf("wrapperTiffReadImage: unsupported byte encoding type.\n");
        return 2;
    }
    if ( (int)width != imageWidth || (int)length != imageHeight) {
        TIFFClose(holoImage);
        printf("wrapperTiffReadImage: input sizes different from the ones of the image.\n");
        return 2;
    }
    
    tsize_t size = TIFFStripSize(holoImage);
    tsize_t imOffset = 0;
    tdata_t buf = _TIFFmalloc(TIFFStripSize(holoImage));
    tsize_t bufferSize = TIFFNumberOfStrips(holoImage) * size;
    unsigned char* auxBuffer = NULL;
    if (bitsPerPixel == 8)
        auxBuffer = (unsigned char*)malloc(bufferSize*sizeof(unsigned char));
    
    for (tstrip_t strip = 0; strip < TIFFNumberOfStrips(holoImage); strip++) {
        size = TIFFReadEncodedStrip(holoImage, strip, buf, (tsize_t)-1); //-1, so that the function calculates the appropriate size
        //Transferring to main buffer
        if (bitsPerPixel == 8) {
            memcpy(auxBuffer+imOffset, buf, size);
            imOffset += size;
        }
        else {
            memcpy(imData+imOffset, buf, size);
            imOffset += size/2; //2 bytes per pixel
        }
    }
    _TIFFfree(buf);
    
    if (bitsPerPixel == 8) {
        //Copying to out buffer
        for (unsigned int k = 0; k < bufferSize; k++)
            imData[k] = (unsigned short) auxBuffer[k];
    }
    
    TIFFClose(holoImage);
    free(auxBuffer);
    
    return 0;
}

int tiffWriteImage(char* filename, int imageHeight, int imageWidth,
                   unsigned char bitsPerPixel, unsigned short* imData) {
    if (filename == NULL || imData == NULL) {
        printf("wrapperTiffWriteImage: null input pointer.\n");
        return 1;
    }
    if (imageHeight < 1 || imageWidth < 1) {
        printf("wrapperTiffWriteImage: width and height must be > 1.\n");
        return 2;
    }
    if (bitsPerPixel != 8 && bitsPerPixel != 16) {
        printf("wrapperTiffWriteImage: unsupported bits per pixel.\n");
        return 2;
    }
    
    TIFFSetWarningHandler(NULL);
    TIFFSetErrorHandler(NULL);
    
    TIFF *image;
    
    // Open the TIFF file
    if((image = TIFFOpen(filename, "w")) == NULL){
        printf("wrapperTiffWriteImage: Could not open file for writing.\n");
        return 2;
    }
    
    // We need to set some values for basic tags before we can add any data
    int status = TIFFSetField(image, TIFFTAG_IMAGEWIDTH, imageWidth);
    if (status)
        status = TIFFSetField(image, TIFFTAG_IMAGELENGTH, imageHeight);
    if (status)
        status = TIFFSetField(image, TIFFTAG_BITSPERSAMPLE, bitsPerPixel);
    if (status)
        status = TIFFSetField(image, TIFFTAG_SAMPLESPERPIXEL, 1);
    if (status)
        status = TIFFSetField(image, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    if (status)
        status = TIFFSetField(image, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
    if (status)
        status = TIFFSetField(image, TIFFTAG_FILLORDER, FILLORDER_MSB2LSB);
    if (status)
        status = TIFFSetField(image, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    
    if (status)
        status = TIFFSetField(image, TIFFTAG_XRESOLUTION, 1.);
    if (status)
        status = TIFFSetField(image, TIFFTAG_YRESOLUTION, 1.);
    if (status)
        status = TIFFSetField(image, TIFFTAG_RESOLUTIONUNIT, RESUNIT_NONE);
    if (!status) {
        TIFFClose(image);
        printf("wrapperTiffWriteImage: error while setting field in tiff writing");
        return 2;
    }
    
    tsize_t size = TIFFDefaultStripSize(image, -1);
    if (!TIFFSetField(image, TIFFTAG_ROWSPERSTRIP, size)) {
        printf("wrapperTiffWriteImage: error while setting field in tiff writing.\n");
        return 2;
    }
    tsize_t writeBytes, imOffset = 0;
    int strip = 0;
    
    unsigned char* auxBuffer = NULL;
    size_t bufferSize = 0;
    if (bitsPerPixel == 8) {
        auxBuffer = (unsigned char*)malloc(imageWidth*imageHeight*sizeof(unsigned char));
        bufferSize = size*imageWidth*sizeof(unsigned char);
        for (int k = 0; k < imageWidth*imageHeight; k++)  //Copy
            auxBuffer[k] = (unsigned char) (imData[k]);
    }
    else
        bufferSize = size*imageWidth*sizeof(unsigned short);
    
    while (imOffset < imageWidth*imageHeight) {
        if (bitsPerPixel == 8) {
            writeBytes = TIFFWriteEncodedStrip(image, strip, auxBuffer + imOffset, bufferSize);
            imOffset += writeBytes;
        }
        else {
            writeBytes = TIFFWriteEncodedStrip(image, strip, imData + imOffset, bufferSize);
            imOffset += writeBytes/2;
        }
        
        strip++;
    }
    if (bitsPerPixel == 8)
        free(auxBuffer);
    
    // Close the file
    TIFFClose(image);
    
    return 0;
}
#else
int csvGetImageSize(char* filename, int* imageHeight, int* imageWidth) {
    if (filename == NULL || imageHeight == NULL || imageWidth == NULL) {
        printf("csvTiffWriteImage: null input pointer.\n");
        return 1;
    }
    FILE* fid = fopen(filename, "r");
    if (fid == NULL) {
        printf("csvTiffWriteImage: impossible to open file %s.\n", filename);
        return 2;
    }
    
    *imageWidth = 0;
    char line[LINELENGTH];
    int cnt1, cnt2 = 0;
    char *ptr;
    //Read first line to check size
    while (fgets(line, LINELENGTH, fid) != NULL) {
        if (strlen(line)+1 == LINELENGTH) {
            fclose(fid);
            printf("csvGetImageSize: internal buffer too small.\n");
            return 2;
        }
        //Check points
        ptr = strstr(line, ".");
        if (ptr != NULL) {
            fclose(fid);
            printf("csvGetImageSize: only integers are allowed.\n");
            return 2;
        }
        
        //Count elements in line
        cnt1 = 0;
        ptr = strstr(line, ",");
        while (ptr - line < strlen(line)) {
            cnt1++;
            ptr = strstr(ptr+1, ",");
        }
        if (*imageWidth == 0) //Set width at first pass
            *imageWidth = cnt1+1;
        else {
            if (cnt1 + 1 != *imageWidth) {
                fclose(fid);
                printf("csvGetImageSize: inconsistent amount of line.\n");
                return 2;
            }
        }
        cnt2++;
    }
    
    *imageHeight = cnt2;
    fclose(fid);
    
    return 0;
}

int csvReadImage(char* filename, int imageHeight, int imageWidth, unsigned short* imData) {
    
    char line[LINELENGTH];
    int k, l;
    char *ptr1, *ptr2;
    char number[10];
    int val;
    
    if (filename == NULL || imData == NULL) {
        printf("csvReadImage: null input pointer.\n");
        return 1;
    }
    FILE* fid = fopen(filename, "r");
    if (fid == NULL) {
        printf("csvReadImage: impossible to open file %s.\n", filename);
        return 2;
    }
    
    for (k = 0; k < imageHeight; k++) {
        ptr1 = fgets(line, LINELENGTH, fid);
        if (ptr1 == NULL) {
            fclose(fid);
            printf("csvReadImage: reached end of file before filling columns.\n");
            return 2;
        }
        if (strlen(line)+1 == LINELENGTH) {
            fclose(fid);
            printf("csvGetImageSize: internal buffer too small.\n");
            return 2;
        }
        //Check points
        ptr1 = strstr(line, ".");
        if (ptr1 != NULL) {
            fclose(fid);
            printf("csvGetImageSize: only integers are allowed.\n");
            return 2;
        }
        
        //Scan line
        ptr1 = line;
        ptr2 = strstr(ptr1, ",");
        if (ptr2 == NULL) {
            fclose(fid);
            printf("csvReadImage: could not find any comma. Incorrect file format.\n");
            return 2;
        }
        strncpy(number, ptr1, ptr2-ptr1);
        number[ptr2-ptr1] = '\0';
        val = atoi(number);
        if (val < 0 || val > 65535) {
            fclose(fid);
            printf("csvReadImage: only values from 0 to 65535 are supported.\n");
            return 2;
        }
        imData[k*imageWidth] = (unsigned short)val;
        ptr1 = ptr2+1;
        for (l = 1; l < imageWidth-1; l++) {
            ptr2 = strstr(ptr1+1, ",");
            if (ptr2 == NULL) {
                fclose(fid);
                printf("csvReadImage: reached end of line before filling line.\n");
                return 2;
            }
            strncpy(number, ptr1, ptr2-ptr1);
            number[ptr2-ptr1] = '\0';
            val = atoi(number);
            if (val < 0 || val > 65535) {
                fclose(fid);
                printf("csvReadImage: only values from 0 to 65535 are supported.\n");
                return 2;
            }
            imData[k*imageWidth + l] = (unsigned short)val;
            ptr1 = ptr2+1;
        }
        //Fill last one
        imData[k*imageWidth + imageWidth - 1] = atoi(ptr1);
    }
    
    fclose(fid);
    return 0;
}

int csvWriteImage(char* filename, int imageHeight, int imageWidth, unsigned short* imData) {
    int k, l;
    
    if (filename == NULL || imData == NULL) {
        printf("csvWriteImage: null input pointer.\n");
        return 1;
    }
    FILE* fid = fopen(filename, "w");
    if (fid == NULL) {
        printf("csvWriteImage: impossible to open file %s.\n", filename);
        return 2;
    }
    
    for (k = 0; k < imageHeight; k++) {
        for (l = 0; l < imageWidth-1; l++)
            fprintf(fid, "%d, ", imData[k*imageWidth + l]);
        fprintf(fid, "%d\n", imData[k*imageWidth + imageWidth - 1]);
    }
    
    fclose(fid);
    return 0;
}
#endif

#ifndef USEFFTW
int checkPow2(int dataSize) {
    // make sure that n is a power of 2
    size_t k = 1;
    size_t logn = 0;
    
    while (k < dataSize) {
        k <<= 1;
        logn++;
    }
    int ntest = (1 << logn) ;
    
    if (dataSize != ntest ) {
        printf("intern2DFFT: matrix is not a power of 2.\n");
        return 1;
    }
    
    return 0;
}

int intern2DFFT(fftw_complex* data, int dataSize, int sign) {
    if (data == NULL) {
        printf("intern2DFFT: null pointer.\n");
        return 1;
    }
    size_t dual;
    size_t bit;
    size_t logn = 0;
    
    // make sure that n is a power of 2
    size_t k = 1;
    while (k < dataSize) {
        k <<= 1;
        logn++;
    }
    int ntest = (1 << logn) ;
    
    if (dataSize != ntest ) {
        printf("intern2DFFT: matrix is not a power of 2.\n");
        return 1;
    }
    
    // bit reverse the ordering of input data for decimation in time algorithm
    bitreverse_order2D(data, dataSize);
    
    // apply fft recursion
    size_t n;
    size_t delta1 = 1, delta2 = dataSize;
    
    for (n = 0; n < 2; n++) {	//Scanning on dimensions
        dual = 1;
        for (bit = 0; bit < logn; bit++) {
            double w_real = 1.0;
            double w_imag = 0.0;
            
            const double theta = 2.0 * sign * M_PI / (2.0 * (double) dual);
            
            const double s = sin (theta);
            const double t = sin (theta / 2.0);
            const double s2 = 2.0 * t * t;
            
            size_t a, b, c;
            //a scans through the different elements within a 2^n block
            //b scans through the various blocks of 2^n size
            //c scans through the various blocks of 2^n size on the second dimension
            
            // a = 0	At a=0, theta is 0, so that the imaginary part of W^k is 0, simplifying calculations
            for (b = 0; b < dataSize; b += 2 * dual) {
                const size_t i = b*delta1;			//i < N/2
                const size_t j = (b+dual)*delta1;	//j > N/2
                for (c = 0; c < dataSize; c++) {	//Applying trigonometric values to each line
                    const size_t i2 = c*delta2 + i;
                    const size_t j2 = c*delta2 + j;
                    
                    //At first position, i=0, j=1
                    const double wd_real = REAL(data,j2);
                    const double wd_imag = IMAG(data,j2);
                    
                    //j >= N/2: Xk = Even_k - W_k * Odd_k
                    IMAG(data,j2) = IMAG(data,i2) - wd_imag;	//w_real = 1, w_imag = 0
                    REAL(data,j2) = REAL(data,i2) - wd_real;
                    //i < N/2: Xk = Even_k + W_k * Odd_k
                    REAL(data,i2) += wd_real;				//w_real = 1, w_imag = 0
                    IMAG(data,i2) += wd_imag;
                }
            }
            
            // a = 1 .. (dual-1)
            for (a = 1; a < dual; a++) {
                // trignometric recurrence for w-> exp(i theta) w
                const double tmp_real = w_real - s * w_imag - s2 * w_real;
                const double tmp_imag = w_imag + s * w_real - s2 * w_imag;
                w_real = tmp_real;
                w_imag = tmp_imag;
                
                for (b = 0; b < dataSize; b += 2 * dual) {
                    const size_t i = (b + a)*delta1;
                    const size_t j = (b + a + dual)*delta1;
                    for (c = 0; c < dataSize; c++) {
                        const size_t i2 = c*delta2 + i;
                        const size_t j2 = c*delta2 + j;
                        
                        //Pre-computing Wk * Odd_k
                        const double wd_real = w_real * REAL(data,j2) - w_imag * IMAG(data,j2);
                        const double wd_imag = w_real * IMAG(data,j2) + w_imag * REAL(data,j2);
                        
                        //j >= N/2: Xk = Even_k - W_k * Odd_k
                        REAL(data,j2) = REAL(data,i2) - wd_real;
                        IMAG(data,j2) = IMAG(data,i2) - wd_imag;
                        //i < N/2: Xk = Even_k + W_k * Odd_k
                        REAL(data,i2) += wd_real;
                        IMAG(data,i2) += wd_imag;
                    }
                }
            }
            dual *= 2;
        }
        //Inverting dimensions
        delta1 = dataSize;
        delta2 = 1;
    }
    
    return 0;	
}

int bitreverse_order2D(fftw_complex* data, int dataSize) {
    if (data == NULL) {
        printf("bitreverse_order2D: null pointer.\n");
        return 1;
    }
    
    // This is the Goldrader bit-reversal algorithm extended for 2D
    size_t n;
    size_t delta1 = 1, delta2 = dataSize;
    
    for (n = 0; n < 2; n++) {	//Scanning on dimensions
        size_t i, j = 0;
        for (i = 0; i < dataSize - 1; i++) {
            size_t k, m = dataSize / 2 ;
            
            if (i < j) {
                for (k = 0; k < dataSize; k++) {	//Scanning on all lines to swap them
                    const size_t i2 = k*delta2 + i*delta1;
                    const size_t j2 = k*delta2 + j*delta1;
                    const double tmp_real = REAL(data,i2);
                    const double tmp_imag = IMAG(data,i2);
                    REAL(data,i2) = REAL(data,j2);
                    IMAG(data,i2) = IMAG(data,j2);
                    REAL(data,j2) = tmp_real;
                    IMAG(data,j2) = tmp_imag;
                }
            }
            
            while (m <= j) {
                j -= m;
                m = m / 2;
            }
            
            j += m ;
        }
        //Inverting dimensions
        delta1 = dataSize;
        delta2 = 1;
    }
    
    return 0;
}

#endif
